import os
import logging
from argparse import ArgumentParser
from bda.recipe.deployment import env
from bda.recipe.deployment.common import (
    DeploymentError,
    Config,
    PackageVersion,
    PWDManager,
    DeploymentPackage,
)

log = logging.getLogger('bda.recipe.deployment')

config = Config(env.CONFIG_PATH)

#-------------------------------------------------------------------------------

epilog =  'Current configured environment: '
epilog += (config.check_env('dev') and 'DEVELOPMENT') or \
          (config.check_env('rc') and 'RELEASE CANDIDATE') or \
          'NOT SET!'

#-------------------------------------------------------------------------------

deployparser = ArgumentParser(description='BDA Deployment Process',
                              epilog=epilog)
deploy_subparsers = deployparser.add_subparsers(help='commands')

def deploy():
    args = deployparser.parse_args()
    args.func(args)

#-------------------------------------------------------------------------------

singleparser = ArgumentParser(description='BDA Deployment Process: '
                                          'Helper Scripts',
                              epilog=epilog)
single_subparsers = singleparser.add_subparsers(help='commands')

def deploy_single():
    args = singleparser.parse_args()
    args.func(args)

#-------------------------------------------------------------------------------

def repopasswd(arguments):
    log.info("Set user and password for server")
    pwdmgr = PWDManager(arguments.distserver[0])
    pwdmgr.set()

sub_rp = deploy_subparsers.add_parser('repopasswd',
                               help='Set user and password for server')
sub_rp.add_argument('distserver', nargs=1,
                   choices=[k for k,v in config.config.items('distserver')],
                   help='name of the distserver')
sub_rp.set_defaults(func=repopasswd)

#-------------------------------------------------------------------------------

def _set_version(package, version):
    log.info("Set version for package")
    deploymentpackage = DeploymentPackage(config, package)
    path = os.path.join(config.sources_dir, package, 'setup.py')
    if not os.path.exists(path):
        log.error("Invalid package name %s" % package)
    pv = PackageVersion(path)
    pv.version = version

def _show_version(package):
    deploymentpackage = DeploymentPackage(config, package)
    log.info("Version: %s = %s" % (package, deploymentpackage.version))

def version(args):
    if config.check_env('dev') and args.version[0] is not 'show':
        return _set_version(args.package[0], args.version[0])
    return _show_version(args.package[0])

if config.check_env('dev'):
    sub_ver = deploy_subparsers.add_parser('version',
                                      help='Shows or sets version of package')
    sub_ver.add_argument('package', nargs=1, help='name of package')
    sub_ver.add_argument('version', nargs='?', help='new version',
                           default='show')
    sub_ver.set_defaults(func=version)
if config.check_env('rc'):
    sub_ver = deploy_subparsers.add_parser('version',
                                      help='Shows version of package')
    sub_ver.add_argument('package', nargs=1, help='name of package')
    sub_ver.set_defaults(func=version)

#-------------------------------------------------------------------------------

def info(args):
    packages = sorted(config.as_dict('packages').keys())
    maxlen = max([len(_) for _ in packages])
    def fill(msg, ml=maxlen):
        return "%s%s" % (msg, ' ' * (ml-len(msg)))
    cols = ("%s " * 5)
    log.info(cols % (maxlen*'-', 10*'-', 10*'-', 10*'-', 10*'-'))
    log.info(cols % (fill('package'),
                     fill('repo (%s)' % config.env, 10),
                     fill('live', 10),
                     fill('rc-branch', 10),
                     fill('location', 10),))
    log.info(cols % (maxlen*'-', 10*'-', 10*'-', 10*'-', 10*'-'))
    for package in sorted(config.as_dict('packages').keys(), key=str.lower):
        dp = DeploymentPackage(config, package)
        p_env = dp.package_options['env']
        rc = dp.rc_source and 'yes' or (p_env=='rc' and p_env or '---')
        log.info(cols % (fill(package),
                         fill(dp.version, 10),
                         fill(dp.live_version or 'not set', 10),
                         fill(rc, 10),
                         fill(config.package(package), 10),))
    log.info(cols % (maxlen*'-', 10*'-', 10*'-', 10*'-', 10*'-'))

sub_inf = deploy_subparsers.add_parser('info',
                                  help='Show information about current state.')
sub_inf.set_defaults(func=info)

#-------------------------------------------------------------------------------

def _commit(package, resource, message):
    log.info("Commit resource")
    deploymentpackage = DeploymentPackage(config, package)
    try:
        deploymentpackage.commit(resource, message)
    except DeploymentError, e:
        log.error("Committing failed: %s" % e)
        return
    except Exception, e:
        log.error("An error occured: %s" % e)

def commit(args):
    return _commit(args.package[0], args.resource[0], args.message[0])

sub_ci = single_subparsers.add_parser('commit',
                                      help='Commit (and push) a resource.')
sub_ci.add_argument('package', nargs=1, help='name of package')
sub_ci.add_argument('resource', nargs='?', help='path to resource')
sub_ci.add_argument('message', nargs='?', help='commit message')
sub_ci.set_defaults(func=commit, resource=None,
                    message='manual deployment commit')

#-------------------------------------------------------------------------------

def merge(cargs):
    log.info("Merge resource to RC")
    deploymentpackage = DeploymentPackage(config, cargs.package[0])
    resource = None
    if cargs.resource:
        resource = cargs.resource[0]
    try:
        deploymentpackage.merge(resource)
    except DeploymentError, e:
        log.error("Merging failed: %s" % e)
        return
    except Exception, e:
        log.error("An error occured: %s" % e)

if config.check_env('rc'):
    sub_merge = deploy_subparsers.add_parser('merge',
                                             help='Merge resource to RC')
    sub_merge.add_argument('package', nargs=1, help='name of package')
    sub_merge.add_argument('resource', nargs='?', help='path to resource')
    sub_merge.set_defaults(func=merge, resource=None)

#-------------------------------------------------------------------------------

def _creatercbranch(all, packages):
    if all:
        packages = config.as_dict('packages').keys()
    for package in packages:
        deploymentpackage = DeploymentPackage(config, package)
        deploymentpackage.check_env("rc")
        try:
            deploymentpackage.creatercbranch()
        except DeploymentError, e:
            log.error("Creating RC branch failed: %s" % e)
            continue
        except Exception, e:
            log.error("An error occured: %s" % e)

def creatercbranch(args):
    return _creatercbranch(args.all, args.package)

if config.check_env('dev'):
    sub_crc = single_subparsers.add_parser('creatercbranch',
                                    help='Create RC branch for one or more '
                                         'or all managed packages.')
    sub_crc_group = sub_crc.add_mutually_exclusive_group()
    sub_crc_group.add_argument('--all', '-a', action='store_true',
                               help='all managed packages')
    sub_crc_group.add_argument('--package', '-p', nargs='+', help='package(s)')
    sub_crc.set_defaults(func=creatercbranch)

#-------------------------------------------------------------------------------

def tag(args):
    log.info("Tag package")
    deploymentpackage = DeploymentPackage(config, args.package[0])
    deploymentpackage.check_env(config.env)
    try:
        deploymentpackage.tag()
    except DeploymentError, e:
        log.error("Tagging failed: %s" % e)
        return
    except Exception, e:
        log.error("An error occured: %s" % e)

sub_tag = single_subparsers.add_parser('tag', help='Tag a package')
sub_tag.add_argument('package', nargs=1, help='name of package')
sub_tag.set_defaults(func=tag)

#-------------------------------------------------------------------------------

def release(cargs):
    log.info("Release package")
    deploymentpackage = DeploymentPackage(config, cargs.package[0])
    deploymentpackage.check_env(config.env)
    try:
        deploymentpackage.release()
    except DeploymentError, e:
        log.error("Releasing failed: %s" % e)
        return
    except Exception, e:
        log.error("An error occured: %s" % e)

sub_rel = single_subparsers.add_parser('release', help='Release package')
sub_rel.add_argument('package', nargs=1, help='name of package')
sub_rel.set_defaults(func=release)

#-------------------------------------------------------------------------------

def exportliveversion(*args):
    log.info("Export live version")
    deploymentpackage = DeploymentPackage(config, args.package[0])
    deploymentpackage.check_env(config.env)
    try:
        deploymentpackage.export_version()
    except DeploymentError, e:
        log.error("Exporting failed: %s" % e)
        return
    except Exception, e:
        log.error("An error occured: %s" % e)

if config.check_env('rc'):
    sub_elv = single_subparsers.add_parser('exportliveversion',
                                           help='Export live version cfg file')
    sub_elv.add_argument('package', nargs=1, help='name of package')
    sub_rel.set_defaults(func=exportliveversion)

#-------------------------------------------------------------------------------

def _exportrcsource(_all, packages):
    if _all:
        packages = config.as_dict('packages').keys()
    log.info("Export rc source")
    for package in packages:
        #XXX check env of package match
        deploymentpackage = DeploymentPackage(config, package)
        try:
            deploymentpackage.export_rc()
        except DeploymentError, e:
            log.error("Exporting failed: %s" % e)
            continue
        except Exception, e:
            log.error("An error occured: %s" % e)

def exportrcsource(args):
    return _exportrcsource(args.all, args.package)

if config.check_env('dev'):
    sub_ers = single_subparsers.add_parser('exportrcsource',
                                           help='Export RC source cfg file')
    sub_ers_group = sub_ers.add_mutually_exclusive_group()
    sub_ers_group.add_argument('--all', '-a', action='store_true',
                               help='all managed packages')
    sub_ers_group.add_argument('--package', '-p', nargs='+', help='package(s)')
    sub_ers.set_defaults(func=exportrcsource)

#-------------------------------------------------------------------------------

def candidate(args):
    """deploy to release candidate

    ./bin/deploy candidate PACKAGENAME VERSIONNUMBER

    - set version
    - commit setup.py
    - create rc branch if not exist
    - export rc sources
    - commit rc sources
    """
    package, newversion = args.package[0], args.version[0]
    deploymentpackage = DeploymentPackage(config, package)
    deploymentpackage.check_env("rc")
    log.info("Complete deployment of release candidate %s with version %s" %
             (package, newversion))
    try:
        _set_version(package, newversion)
        deploymentpackage.commit('setup.py', 'Version Change')
        deploymentpackage.creatercbranch()
        deploymentpackage.export_rc()
        deploymentpackage.commit_rc_source()
    except DeploymentError, e:
        log.error("Candidate deployment failed: %s" % e)
    except Exception, e:
        log.error("An error occured: %s" % e)
        raise

if config.check_env('dev'):
    sub_can = deploy_subparsers.add_parser('candidate',
                                    help='Deploy package to release candidate')
    sub_can.add_argument('package', nargs=1, help='name of package')
    sub_can.add_argument('version', nargs=1, help='new version number')
    sub_can.set_defaults(func=candidate)

#-------------------------------------------------------------------------------

def fullrelease(args):
    """deploy to release on package index

    ./bin/deploy release [packagename]

    - tag version
    - export live version
    - release to package index server
    - commit live versions
    """
    package = args.package[0]
    deploymentpackage = DeploymentPackage(config, package)
    deploymentpackage.check_env(config.env)
    log.info("Complete deployment of final package %s" % package)
    try:
        deploymentpackage.tag()
    except DeploymentError, e:
        log.error("Tagging failed: %s" % e)
        return
    except Exception, e:
        log.error("An error occured: %s" % e)
    try:
        deploymentpackage.export_version()
    except DeploymentError, e:
        log.error("Exporting failed: %s" % e)
        return
    except Exception, e:
        log.error("An error occured: %s" % e)
    try:
        deploymentpackage.release()
    except DeploymentError, e:
        log.error("Releasing failed: %s" % e)
        return
    except Exception, e:
        log.error("An error occured: %s" % e)
    try:
        deploymentpackage.commit_versions()
    except Exception, e:
        log.error("An error occured: %s" % e)

sub_rls = deploy_subparsers.add_parser('release',
                                help='Deploy release to package index')
sub_rls.add_argument('package', nargs=1, help='name of package')
sub_rls.set_defaults(func=fullrelease)



# -*- coding: utf-8 -*-
import os
import sys
import re
import copy
import getpass
import ConfigParser
import logging
from distutils.dist import Distribution
from bda.recipe.deployment import env

log = logging.getLogger('bda.recipe.deployment')

version_pattern = re.compile("""[ \t]*version[ \t]*=[ \t]*["\'](.*)["\'].*""")


class DeploymentError(Exception): pass


class _ConfigMixin(object):

    def __init__(self, path):
        self.path = path
        self.config = ConfigParser.ConfigParser()
        self.config.optionxform = str
        if os.path.exists(path):
            self.config.read(path)

    def __call__(self):
        file = open(self.path, 'wb')
        self.config.write(file)
        file.close()

    def as_dict(self, section):
        return dict(self.config.items(section))

    def read_option(self, section, name):
        if self.config.has_option(section, name):
            return self.config.get(section, name)


class Config(_ConfigMixin):

    def __init__(self, path, buildout_base=None, distserver=None, packages=None,
                 sources=None, rcsources=None, rcversions=None, live=None,
                 env=None, sources_dir=None, register=None):
        _ConfigMixin.__init__(self, path)
        self.packages = packages
        if not self.config.has_section('distserver'):
            self.config.add_section('distserver')
        if not self.config.has_section('packages'):
            self.config.add_section('packages')
        if not self.config.has_section('sources'):
            self.config.add_section('sources')
        if not self.config.has_section('settings'):
            self.config.add_section('settings')
        if distserver is not None:
            for key, val in distserver.items():
                self.config.set('distserver', key, val)
        if packages is not None:
            for key, val in packages.items():
                self.config.set('packages', key, val)
        if sources is not None:
            for key, val in sources.items():
                self.config.set('sources', key, val)
        if buildout_base is not None:
            self.config.set('settings', 'buildout_base', buildout_base)
        if rcsources is not None:
            self.config.set('settings', 'rc_sources', rcsources)
        if rcversions is not None:
            self.config.set('settings', 'rc_versions', rcversions)
        if live is not None:
            self.config.set('settings', 'live_versions', live)
        if env is not None:
            self.config.set('settings', 'env', env)
        if sources_dir is not None:
            self.config.set('settings', 'sources_dir', sources_dir)
        if register is not None:
            self.config.set('settings', 'register', register)

    @property
    def buildout_base(self):
        return self.read_option('settings', 'buildout_base')

    @property
    def rc_versions(self):
        return self.read_option('settings', 'rc_versions')

    @property
    def rc_sources(self):
        return self.read_option('settings', 'rc_sources')

    @property
    def live_versions(self):
        return self.read_option('settings', 'live_versions')

    @property
    def env(self):
        return self.read_option('settings', 'env')

    @property
    def sources_dir(self):
        return self.read_option('settings', 'sources_dir')

    @property
    def registerdist(self):
        return self.read_option('settings', 'register')

    def distserver(self, name):
        return self.read_option('distserver', name)

    def _package_split(self, pkgstr):
        if not pkgstr:
            return None, {}
        parts = pkgstr.strip().split()
        options = dict()
        options.setdefault('env', 'rc')
        if len(parts) > 1:
            for optionstr in parts[1].split(','):
                key, value = optionstr.split('=')
                options[key] = value
        return parts[0], options

    def package(self, name):
        return self._package_split(self.read_option('packages', name))[0]

    def package_options(self, name):
        return self._package_split(self.read_option('packages', name))[1]

    def source(self, name):
        return self.read_option('sources', name)

    def check_env(self, env):
        return self.env in ['all', env]


class RcSourcesCFG(_ConfigMixin):

    def __init__(self, path):
        _ConfigMixin.__init__(self, path)
        if not self.config.has_section('sources'):
            self.config.add_section('sources')

    def set(self, package, source):
        self.config.set('sources', package, source)

    def get(self, package):
        return self.read_option('sources', package)


class VersionsCFG(_ConfigMixin):

    def __init__(self, path):
        _ConfigMixin.__init__(self, path)
        if not self.config.has_section('versions'):
            self.config.add_section('versions')

    def set(self, package, version):
        self.config.set('versions', package, version)

    def get(self, package):
        return self.config.get('versions', package)


class ReleaseCFG(_ConfigMixin):

    def set(self, server, user, password):
        if not self.config.has_section(server):
            self.config.add_section(server)
        self.config.set(server, 'username', user)
        self.config.set(server, 'password', password)

    def get(self, server):
        if not self.config.has_option(server, 'username') \
          or not self.config.has_option(server, 'password'):
            return None
        return self.config.get(server, 'username'), \
               self.config.get(server, 'password')


class PackageVersion(object):

    def __init__(self, path):
        self.path = path

    def _get_version(self):
        file = open(self.path)
        version = "0"
        for line in file.readlines():
            mo = version_pattern.match(line)
            if mo:
                version = mo.group(1)
                break
        file.close()
        return version

    def _set_version(self, value):
        out = list()
        file = open(self.path)
        for line in file.readlines():
            mo = version_pattern.match(line)
            if mo:
                line = line[:mo.span(1)[0]] + value + line[mo.span(1)[1]:]
            out.append(line)
        file.close()
        file = open(self.path, 'w')
        file.writelines(out)
        file.close()

    version = property(_get_version, _set_version)


class PWDManager(object):

    def __init__(self, server):
        self.server = server
        self.releaserc = ReleaseCFG(env.CFG_PATH)

    def get(self):
        res = self.releaserc.get(self.server)
        if res is not None:
            return res
        self.set()
        return self.releaserc.get(self.server)

    def set(self):
        username = password = None
        while not username:
            username = raw_input('Username: ')
        while not password:
            password = getpass.getpass('Password: ')
        self.releaserc.set(self.server, username, password)
        self.releaserc()


class DeploymentPackage(object):

    connectors = dict()

    def __init__(self, config, package):
        self.config = config
        self.package = package

    def check_env(self, target_env):
        if self.package_options['env'] == target_env:
            return True
        raise DeploymentError(
                "action for package '%s' for target env '%s' is not allowed (allow=%s)." %
                (self.package, target_env, self.package_options['env']))

    def commit(self, resource, message):
        """Commit resource of package with message.

        @param resource: path to resource. If None, all resources in package
                         are committed
        @param message: commit message
        """
        self.connector.commit(resource, message)

    def commit_buildout(self, resource, message):
        """Commit resource of package with message.

        @param resource: path to resource. If None, all resources in package
                         are committed
        @param message: commit message
        """
        self.connector.commit_buildout(resource, message)

    def commit_rc_source(self):
        """Function committing RC source file.
        """
        self.commit_buildout(self.config.rc_sources, '"RC Sources changed"')

    def commit_versions(self):
        """Function committing LIVE/RC source file.
        """
        if self.package_options['env'] == 'dev':
            self.commit_buildout(self.config.rc_versions, '"RC versions updated"')
        self.commit_buildout(self.config.live_versions, '"LIVE versions updated"')

    def merge(self, resource=None):
        """Merge from trunk to rc.

        Function only callable in ``rc`` environment.

        Raise ``DeploymentError`` if called in wrong environment.

        @param resource: path to resource. If None, all resources in package
                         are merged
        """
        self.connector.merge(resource)

    def creatercbranch(self):
        """Create RC branch for package.
        """
        self.connector.creatercbranch()

    def tag(self):
        """Tag package from rc to tags/version. Use version of
        package ``setup.py``

        Function only callable in ``rc`` environment.

        Raise ``DeploymentError`` if tag already exists or if called in
        wrong environment.
        """
        self.connector.tag()

    def release(self):
        """Release package to configured dist server.

        Raise ``DeploymentError`` if called in wrong environment.

        XXX: make me thread safe.
        """
        setup = os.path.join(self.package_path, 'setup.py')
        old_argv = copy.copy(sys.argv)
        os.chdir(self.package_path)
        setup_globals = copy.copy(globals())
        setup_globals['__file__'] = setup
        if self.dist_server.startswith('file://'):
            sys.argv = ['setup.py',
                        'sdist',
                        '-d',
                        self.dist_server[7:]]
            res = execfile('setup.py', setup_globals)
        else:
            pwdmgr = PWDManager(self.config.package(self.package))
            username, password = pwdmgr.get()
            sys.argv = ['setup.py',
                        'sdist',
                        'deploymentregister',
                        'deploymentupload']
            if self.config.package(self.package) in self.register_dist:
                sys.argv.append('deploymentregister')
            env.waitress = {
                'repository': self.dist_server,
                'username': username,
                'password': password,
            }
            res = execfile('setup.py', setup_globals)
            env.waitress = dict()
        sys.argv = old_argv

    def export_rc(self):
        """Export package rc repo info to configured rc sources config.

        Function only callable in ``dev`` environment.
        """
        sources = RcSourcesCFG(self.config.rc_sources)
        sources.set(self.package, self.connector.rc_source)
        sources()

    @property
    def rc_source(self):
        sources = RcSourcesCFG(self.config.rc_sources)
        return sources.get(self.package)

    def export_version(self):
        """Export current resource version to configured live/rc versions
        config.
        """
        if self.package_options['env'] == 'dev':
            versions = VersionsCFG(self.config.rc_versions)
            versions.set(self.package, self.version)
            versions()
        versions = VersionsCFG(self.config.live_versions)
        versions.set(self.package, self.version)
        versions()

    @property
    def rc_version(self):
        versions = VersionsCFG(self.config.rc_versions)
        return versions.read_option('versions', self.package)

    @property
    def live_version(self):
        versions = VersionsCFG(self.config.live_versions)
        return versions.read_option('versions', self.package)

    @property
    def _source(self):
        source = self.config.source(self.package)
        if source is None:
            raise KeyError, \
                  'no package %s found in [sources] section!' % self.package + \
                  ' maybe misspelled?'
        return source

    @property
    def connector_name(self):
        return self._source.split(' ')[0]

    @property
    def connector(self):
        return self.connectors[self.connector_name](self)

    @property
    def package_path(self):
        return os.path.join(self.config.sources_dir, self.package)

    @property
    def package_options(self):
        return self.config.package_options(self.package)

    @property
    def dist_server(self):
        return self.config.distserver(self.config.package(self.package))

    @property
    def register_dist(self):
        return self.config.registerdist

    @property
    def buildout_base(self):
        return self.config.buildout_base

    @property
    def version(self):
        path = os.path.join(self.package_path, 'setup.py')
        if os.path.exists(path):
            return PackageVersion(path).version
        else:
            return 'unversioned'

    @property
    def package_uri(self):
        source = self.config.source(self.package)
        return self._source.split(' ')[1].rstrip('/')



import os
import tempfile
import unittest
import interlude
import zope.component
from pprint import pprint
from zope.testing import doctest
from zope.configuration.xmlconfig import XMLConfig

optionflags = doctest.NORMALIZE_WHITESPACE | \
              doctest.ELLIPSIS | \
              doctest.REPORT_ONLY_FIRST_FAILURE

TESTFILES = [
    'common.rst',
    'svn.rst',
    'git.rst',
    'main.rst',
]

tempdir = tempfile.mkdtemp()


def test_suite():
    XMLConfig('meta.zcml', zope.component)()
    return unittest.TestSuite([
        doctest.DocFileSuite(
            file, 
            optionflags=optionflags,
            globs={
                'pprint': pprint,
                'tempdir': tempdir,
                'interact': interlude.interact
            },
        ) for file in TESTFILES
    ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
import zc.buildout
import shutil
import os
from bda.recipe.deployment.common import Config


class Recipe(object):

    def __init__(self, buildout, name, options):
        self.name = name
        self.options = options
        self.buildout = buildout
        self.buildout_base = buildout['buildout']['directory']
        distserver = options.get('distserver')
        if not distserver:
            raise zc.buildout.UserError(u'distserver section missing.')
        distserver = [d.strip() for d in distserver.strip().split('\n')]
        if not distserver:
            raise zc.buildout.UserError(u'No dist servers defined.')
        self.distserver = dict()
        for server in distserver:
            key, val = server.split(' ')
            self.distserver[key] = val
        packages = options.get('packages')
        if not packages:
            raise zc.buildout.UserError(u'packages section missing.')
        packages = [p.strip() for p in packages.strip().split('\n')]
        if not packages:
            raise zc.buildout.UserError(u'No packages defined.')
        self.packages = dict()
        for package in packages:
            key, val = package.split(None, 1)
            self.packages[key] = val
        base_path = buildout['buildout']['directory']
        
        self.dev_sources = options.get('dev_sources')
        if not self.dev_sources:
            raise zc.buildout.UserError(u'No DEV sources config defined.')
        if not self.dev_sources.startswith(base_path):
            self.dev = os.path.join(base_path, self.dev_sources)
        
        self.rc_sources = options.get('rc_sources')
        if not self.rc_sources:
            raise zc.buildout.UserError(u'No RC sources config defined.')
        if not self.rc_sources.startswith(base_path):
            self.rc_sources = os.path.join(base_path, self.rc_sources)
            
        self.rc_versions = options.get('rc_versions')
        if not self.rc_versions:
            raise zc.buildout.UserError(u'No RC versions config defined.')
        if not self.rc_versions.startswith(base_path):
            rc = os.path.join(base_path, self.rc_versions)
        
        self.live_versions = options.get('live_versions')
        if not self.live_versions:
            raise zc.buildout.UserError(u'No Live versions config defined.')
        if not self.live_versions.startswith(base_path):
            self.live_versions = os.path.join(base_path, self.live_versions)
        
        self.register = options.get('register', '')
        self.env = options.get('env')
        if not self.env in ['dev', 'rc', 'all']:
            raise zc.buildout.UserError(u'No or wrong env flavor defined.')
        sources_default = os.path.join(self.buildout['buildout']['directory'],
                                       'src')
        self.sources_dir = self.buildout['buildout'].get('sources-dir', 
                                                         sources_default) 

    def install(self):
        path = os.path.join(self.buildout['buildout']['directory'],
                            '.bda.recipe.deployment.cfg')
        if os.path.exists(path):
            os.remove(path)
        dev_sources =  Config(self.dev_sources)
        sources = dev_sources.as_dict('sources')
        Config(path, self.buildout_base, self.distserver, self.packages,
               sources, self.rc_sources, self.rc_versions, self.live_versions, 
               self.env, self.sources_dir, self.register)()

    def update(self):
        return self.install()



import sys
import logging

log = logging.getLogger()

log.setLevel(logging.DEBUG)

hdlr = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(levelname)-5s %(message)s')

hdlr.setLevel(logging.INFO)
hdlr.setFormatter(formatter)
log.addHandler(hdlr)

import svn
import git



import os
import subprocess
from mr.developer.git import gitWorkingCopyFactory
from bda.recipe.deployment.common import DeploymentError
from bda.recipe.deployment.common import DeploymentPackage
import logging

log = logging.getLogger('bda.recipe.deployment git')

CLEAN, DIRTY = 'clean', 'dirty'


class GitConnector(object):

    def __init__(self, package):
        self.package = package

    @property
    def rc_source(self):
        return 'git %s branch=rc' % self.package.package_uri

    def source(self, context="package"):
        if context=='package':
            return dict(name=self.package.package, 
                        path=self.package.package_path,
                        url=self.package.package_uri)
        elif context=='buildout':
            return dict(name='buildout', 
                        path=self.package.buildout_base,
                        url=None)
        else: 
            raise DeploymentError('Commit context "%s" not allowed.' % context)

    def git_wc(self, context="package"):
        return gitWorkingCopyFactory(self.source(context=context))

    def _rungit(self, command, msg='', context='package'):
        """runs git command in a given context
        
        command
            list of strings (paramters) after git 
            
        msg
            error message if command fails
        
        context
            string, one of 'package', 'buildout'
        
        """
        wc = self.git_wc(context=context)
        if context=='package':
            cwd = self.package.package_path
        else:
            cwd = self.package.buildout_base
        cmd = wc.run_git(command, cwd=cwd)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            return stdout, stderr, cmd
        log.error(msg)
        command = '%s' % ' '.join(command)
        message = '\n'.join([command, msg, stdout, stderr]) 
        raise DeploymentError('Failed command: %s' % message)

    def _pull(self, branch='master'):
        log.info('Initiate pull origin %s' % branch)
        cmd = ['pull', 'origin', branch]
        stdout, stderr, cmd = self._rungit(cmd)
        log.info('Pull done.')                        

    def commit(self, resource='-a', message='bda.recipe.deployment commit'):
        """Commit means here a commit and push in one
        """
        if self.status == CLEAN:            
            log.info('Abort commit %s, working directory is clean' % \
                     (resource == '-a' and 'all' or resource))
            return
        log.info('Initiate commit  %s' % (resource == '-a' and 'all' or 
                                          resource))
        if resource != '-a':
            stdout, stderr, cmd = self._rungit(["add", resource])
        message = '"%s"' % message
        stdout, stderr, cmd = self._rungit(["commit", resource, '-m', message])
        stdout, stderr, cmd = self._rungit(["push"])
        log.info('Commit done.')                        

    def commit_buildout(self, resource='-a', 
                        message='bda.recipe.deployment buildout commit'):
        if self.status_buildout == CLEAN:            
            log.info('Abort buildout commit  %s, working directory is clean' % \
                     (resource == '-a' and 'all' or resource))
            return
        log.info('Initiate buildout commit  %s' % (resource == '-a' and 'all' or 
                                          resource))
        if resource != '-a':
            stdout, stderr, cmd = self._rungit(["add", resource],
                                               context='buildout')
        stdout, stderr, cmd = self._rungit(["commit", resource, '-m', message],
                                           context='buildout')
        stdout, stderr, cmd = self._rungit(["push"], context='buildout')
        log.info('Commit done.')                        

    def _has_rc_branch(self, remote=False):
        branches = self._get_branches()
        context = remote and 'origin' or None
        return bool([_ for _ in branches 
                     if _['branch']=='rc' and _['remote']==context])

    def _current_branch(self):
        return [_['branch'] for _ in self._get_branches() if _['current']][0]

    def _get_branches(self):
        """a list with value as dict with:
            * key=branch: branch-name
            * key=remote: remote-name or None (for local)
            * key=current: (bool) (only possible for local)
            * key=alias: (string or None) this branch is the HEAD or other alias
        """   
        stdout, stderr, cmd = self._rungit(["branch", "-a"])
        result = list()
        aliases = dict()
        def _location_split(loc):
            if loc.startswith('remotes'):
                prefix, remotename, branchname = loc.split('/')
            elif '/' in loc:
                remotename, branchname = loc.split('/')
            else:
                remotename, branchname = None, loc
            return remotename, branchname
        for line in stdout.split('\n'):
            if not line:
                continue
            current = line.startswith('*')
            line = line[2:]
            if '->' in line:
                key, target = line.split('->')
                key = '/'.join(_location_split(key.strip()))
                aliases[key] = target.strip() 
                continue
            if line.startswith('remotes'):
                prefix, remotename, branchname = line.split('/')
            else:
                remotename, branchname = None, line
            result.append(dict(branch=branchname, 
                               remote=remotename, 
                               current=current,
                               alias=None))
        for alias, target in aliases.items():
            remote, branchname = _location_split(target)
            for item in result:
                if item['branch'] != branchname or item['remote'] != remotename:
                     continue
                item['alias'] = alias                      
        return result     

    def creatercbranch(self):
        """creates rc branch if not exists"""
        log.info('Initiate creation of RC branch')
        # check if clean, if not commit
        if self.status == DIRTY:
            self.commit(message='bda.recipe.deployment pre create rc branch')
        # check if branch already exists, if yes, log and return direct
        if self._has_rc_branch():
            log.warning('RC branch already exist, abort create.')
            # here (not sure) we might check if a remote branch exists and if 
            # yes, connect it to the local branch, but OTOH this can be wrong
            return False
        if self._has_rc_branch(remote=True):
            log.info('Remote rc branch exists, checkout')
            stdout, stderr, cmd = self._rungit(["checkout", "-b", "rc", 
                                                "origin/rc"])
            stdout, stderr, cmd = self._rungit(["checkout", "master"])            
            return True
        else:
            log.info('No remote rc branch, checkout new and push')
            stdout, stderr, cmd = self._rungit(["checkout", "-b", "rc"])
            stdout, stderr, cmd = self._rungit(["push", "-u", "origin", "rc"])
            stdout, stderr, cmd = self._rungit(["checkout", "master"])            
            return True

    def merge(self, resource=None):
        """merges changes from dev branch to rc branch"""
        log.info('Merge master into rc branch')
        if self.status == DIRTY:
            self.commit(message='bda.recipe.deployment pre merge commit')
        if self.status == DIRTY:
            raise DeploymentError('Not clean after pre merge commit: %s' %\
                                  self.package.package)
        if not self._has_rc_branch():
            # hmm, do we need this?
            self.creatercbranch()
        self._pull()
        self._pull('rc')
        stdout, stderr, cmd = self._rungit(["checkout", "rc"])
        stdout, stderr, cmd = self._rungit(["merge", "master"])
        log.info('Merge done')

    def _tags(self):
        stdout, stderr, cmd = self._rungit(["tag"])
        return [_.strip() for _ in stdout.split('\n') if _.strip()]

    def _tag(self, version, msg):
        stdout, stderr, cmd = self._rungit(["tag", '-a', version, '-m', msg])

    def tag(self):
        """Tag package from rc  with version. Use version of
        package ``setup.py``
        """
        version = self.package.version
        log.info('Tag version %s' % version) 
        if self.status == DIRTY:
            self.commit(message='bda.recipe.deployment pre tag commit')
        if version in self._tags():
            raise DeploymentError('Tag %s already exist for %s.' % 
                                  (version, self.package.package))            
        self._tag(version, 'version tag set by bda.recipe.deployment') 
        stdout, stderr, cmd = self._rungit(["push", "--tags"])
        log.info('Tagging done')

    # proxy method
    @property
    def status(self):
        wc = self.git_wc()
        return wc.status()

    @property
    def status_buildout(self):
        wc = self.git_wc(context='buildout')
        return wc.status()

DeploymentPackage.connectors['git'] = GitConnector



import os

BASE_DIR = os.path.abspath('')
if BASE_DIR.endswith('/bin'):
    BASE_DIR = BASE_DIR[:-4]

CFG_PATH = os.path.join(os.environ['HOME'], '.bda.recipe.deployment.auth')

CONFIG_PATH = os.path.join(BASE_DIR, '.bda.recipe.deployment.cfg')

waitress = dict()



from setuptools.command.upload import upload as st_upload
from setuptools.command.register import register as st_register
from bda.recipe.deployment import env


class upload(st_upload):

    def finalize_options(self):
        """Set repository, username and password for uploading by
        bda.recipe.deployment.env.waitress
        """
        self.repository = env.waitress['repository']
        self.username = env.waitress['username']
        self.password = env.waitress['password']        
        st_upload.finalize_options(self)


class register(st_register):

    def finalize_options(self):
        """Set repository, username and password for register by
        bda.recipe.deployment.env.waitress
        """
        self.repository = env.waitress['repository']
        self.username = env.waitress['username']
        self.password = env.waitress['password']
        st_register.finalize_options(self)



from setuptools import setup, find_packages
import os
version = "2.0b5"

shortdesc = "python deployment process based on buildout and mr.developer"
longdesc = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
longdesc += open(os.path.join(os.path.dirname(__file__), 'LICENSE.rst')).read()

setup(name="bda.recipe.deployment",
      version=version,
      description=shortdesc,
      long_description=longdesc,
      classifiers=[
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Topic :: Software Development',
      ],
      keywords="",
      author="BlueDynamics Alliance",
      author_email="dev@bluedynamics.com",
      url="http://pypi.python.org/pypi/bda.recipe.deployment",
      license="GPLv2",
      packages=find_packages("src"),
      package_dir={"": "src"},
      namespace_packages=["bda", "bda.recipe"],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'zope.component',
          'zope.configuration',
          'mr.developer',
      ],
      extras_require=dict(
          test=[
            'interlude',
            'zope.testing',
          ]
      ),
      entry_points="""
      [zc.buildout]
      default = bda.recipe.deployment.recipe:Recipe

      [distutils.commands]
      deploymentregister = bda.recipe.deployment.command:register
      deploymentupload = bda.recipe.deployment.command:upload

      [console_scripts]
      deploy = bda.recipe.deployment.main:deploy
      bda_deployment_helper = bda.recipe.deployment.main:deploy_single
      """,
      )



import os
import subprocess
from mr.developer.svn import SVNWorkingCopy
from bda.recipe.deployment.common import DeploymentError
from bda.recipe.deployment.common import DeploymentPackage
import logging

log = logging.getLogger('bda.recipe.deployment svn')


class SVNConnector(SVNWorkingCopy):

    def __init__(self, package):
        if not package.package_uri.endswith('trunk') \
           and '/branches/' not in package.package_uri:
            msg = u"Invalid repository structure. Could only handle " + \
                  u"packages contained either in 'svn_url/trunk' or " + \
                  u"in 'svn_url/branches/NAME'"
            raise DeploymentError(msg)
        self.package = package
        self.source = dict()
        self.source['name'] = package.package
        self.source['path'] = package.package_path
        self.source['url'] = package.package_uri

    @property
    def rc_source(self):
        return 'svn %s' % self._rc_uri

    def commit(self, resource, message):
        url = self.package.package_uri
        resource = os.path.join(self.package.config.sources_dir,
                                self.package.package,
                                resource)
        args = ["svn", "ci", resource, '-m"%s"' % message]
        kwargs = {}
        msg = ' '.join(args)
        log.info(msg)
        stdout, stderr, returncode = self._svn_communicate(args, url,
                                                           **kwargs)

    def commit_buildout(self, resource, message):
        self.commit(resource, message)

    def merge(self, resource=None):
        """
        svn ci path/to/foo -m 'RC Merge'
        svn merge https://foo/branches/rc https://foo/trunk .
        """
        self.commit('', 'RC Merge')
        from_resource = self.package.package_uri
        to_resource = self._rc_uri
        wc_resource = os.path.join(self.package.config.sources_dir,
                                   self.package.package)
        if resource is not None:
            from_resource = '/'.join([from_resource, resource])
            to_resource = '/'.join([to_resource, resource])
            wc_resource = os.path.join(wc_resource, resource)
        args = ["svn", "merge", to_resource, from_resource, wc_resource]
        log.info(' '.join(args))
        kwargs = {}
        stdout, stderr, returncode = self._svn_communicate(args, from_resource, 
                                                           **kwargs)
        if returncode != 0:
            msg = "Subversion merge for '%s' failed.\n%s" % (resource, stderr)
            raise DeploymentError(msg)
        if kwargs.get('verbose', False):
            return stdout

    def creatercbranch(self):
        source_uri = self.package.package_uri
        branches_path = '%s/branches' % self._svn_base_path
        if not self._svn_exists(branches_path):
            msg = "'Create branches directory for %s'" % self.package.package
            log.info(msg)
            args = ["svn", "mkdir", branches_path, '-m', msg]
            kwargs = {}
            stdout, stderr, returncode = self._svn_communicate(args, source_uri,
                                                               **kwargs)
            if returncode != 0:
                msg = u"'Cannot create directory %s'" % branches_path
                raise DeploymentError(msg)
        if not self._svn_exists(self._rc_uri):
            msg = "'Create RC branch for %s'" % self.package.package
            log.info(msg)
            self._svn_copy(source_uri, self._rc_uri, msg)
        else:
            msg = "'RC branch for %s already exists. Use merge script in " + \
                  "RC environment to synchronize resources.'"
            msg = msg % self.package.package
            log.info(msg)

    def tag(self):
        msg = "'Tag %s version %s'" % (self.package.package,
                                      self.package.version)
        log.info(msg)
        if self._svn_exists(self._tag_uri):
            msg = "Tagging for '%s' failed. Version %s already exists" % (
                        self.package.package, self.package.version)
            raise DeploymentError(msg)
        self.commit('', 'RC Tag')
        tags_path = '%s/tags' % self._svn_base_path
        if not self._svn_exists(tags_path):
            url = self.package.package_uri
            msg = "'Create tags directory for %s'" % self.package.package
            log.info(msg)
            args = ["svn", "mkdir", tags_path, '-m', msg]
            kwargs = {}
            stdout, stderr, returncode = self._svn_communicate(args, url,
                                                               **kwargs)
            if returncode != 0:
                msg = u"'Cannot create directory %s'" % tags_path
                raise DeploymentError(msg)
        msg = "'Tag %s version'" % self.package.version
        self._svn_copy(self.package.package_uri, self._tag_uri, msg)

    def _svn_exists(self, uri):      
        log.info("Check for %s" % uri)
        cmd = subprocess.Popen(["svn", "ls", "--non-interactive", uri],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
        stdout, stderr = cmd.communicate()
        if cmd.returncode != 0:
            log.info("...not found %s" % uri)
            return False
        log.info("...found %s" % uri)
        return True

    def _svn_copy(self, source, target, message):
        url = self.package.package_uri
        args = ["svn", "cp", source, target, '-m', message]
        log.info(' '.join(args))
        kwargs = {}
        stdout, stderr, returncode = self._svn_communicate(args, url, **kwargs)
        if returncode != 0:
            msg = "Subversion copy failed.\n%s -> %s\n%s" % (source,
                                                             target, stderr)
            raise DeploymentError(msg)
        if kwargs.get('verbose', False):
            return stdout

    @property
    def _svn_base_path(self):
        uri = self.package.package_uri
        idx = uri.find('/trunk')
        if idx < 1:
            idx = idx = uri.find('/branches/')
        if idx < 1:
            raise ValueError, \
                  'URI not valid (trunk or branches needed): %s' % uri
        return uri[:idx]

    @property
    def _rc_uri(self):
        uri = '%s/branches/rc' % self._svn_base_path
        if self.package.package_uri.endswith('/trunk'):
            return uri
        idx = self.package.package_uri.rfind('/') + 1
        uri += '-%s' % self.package.package_uri[idx:]
        return uri

    @property
    def _tag_uri(self):
        return '%s/tags/%s' % (self._svn_base_path, self.package.version)

DeploymentPackage.connectors['svn'] = SVNConnector
